==> 1.8.3.zip: entaxy-public/platform/runtime/base/objects-base/object-factory/src/main/java/ru/entaxy/platform/base/objects/factory/impl/GenerationHelper.java <==
/*-
 * ~~~~~~licensing~~~~~~
 * test-producers
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.base.objects.factory.impl;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.osgi.service.component.annotations.CollectionType;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.osgi.service.component.annotations.ReferencePolicy;
import org.osgi.service.component.annotations.ReferencePolicyOption;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.base.generator.template.Template;
import ru.entaxy.base.generator.template.TemplateService;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generated;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generator;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.factory.GeneratorFactory;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.OutputInfo;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.SCOPE;
import ru.entaxy.platform.base.support.CommonUtils;

@Component(immediate = true, service = {GenerationHelper.class})
public class GenerationHelper {

	private static final Logger log = LoggerFactory.getLogger(GenerationHelper.class);

	@Reference(cardinality = ReferenceCardinality.MANDATORY, collectionType = CollectionType.SERVICE, policy = ReferencePolicy.DYNAMIC, policyOption = ReferencePolicyOption.GREEDY)
	protected volatile TemplateService templateService;
	
	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	FactoryRegistry factoryRegistry;
	
	public Generated generateForFactory(EntaxyFactory factory, String outputType, SCOPE scope, Map<String, Object> parameters) throws Exception {
		
		log.debug("GENERATING for factory [" + factory.getId() + "]");
		log.debug("GENERATING for factory [" + factory.getId() + "] :: PARAMETERS \n"
				 + parameters.entrySet().stream().map(e -> e.getKey() + ":" + (null==e.getValue()?"null":e.getValue()))
				 .collect(Collectors.toList())
				 .stream().collect(Collectors.joining(", "))
				);
		
		OutputInfo outputInfo = factory.getOutputByType(outputType);
		
		if (outputInfo == null) {
			// @TODO throw Exception
			log.debug("GENERATING for factory [" + factory.getId() + "] :: output [{}] not found", outputType);
			return null;
		}
			
		if (!outputInfo.getScopes().contains(scope.label)) {
			// @TODO throw Exception
			log.debug("GENERATING for factory [" + factory.getId() + "] :: scope [{}] not supported", scope.label);
			return null;
		}
			
		Map<String, Object> generationProperties = new HashMap<>();
		generationProperties.put(EntaxyFactory.GENERATION.ATTRIBUTES.OUTPUT, outputType);
		generationProperties.put(EntaxyFactory.GENERATION.ATTRIBUTES.SCOPE, scope.label);
		generationProperties.put(EntaxyFactory.GENERATION.ATTRIBUTES.FACTORY, factory.getId());
		
		// try to generate via GeneratorFactory
		log.debug("GENERATING for factory [" + factory.getId() + "] :: TRY via GeneratorFactory");
		
		Map<String, Object> generatorProperties = new HashMap<>(outputInfo.getConfig());
		generatorProperties.putAll(generationProperties);
		Generator g = GeneratorFactory.createGenerator(outputInfo.getGenerator(), generatorProperties);
		
		Map<String, Object> generationParameters = new HashMap<>(parameters);
		generationParameters.putAll(generationProperties);
		
		if (g.isGeneratable(generationParameters)) {
			log.debug("GENERATING for factory [" + factory.getId() + "] :: generating");
			return g.generate(generationParameters);
		} else {
			log.debug("GENERATING for factory [" + factory.getId() + "] :: NOT generatable");			
		}
		
		// try to generate via TemplateService
		log.debug("GENERATING for factory [" + factory.getId() + "] :: TRY via TemplateService");
		List<String> candidates = new LinkedList<>();
		candidates.add(factory.getId() + "." + outputType + "." + scope.label);
		candidates.add(factory.getId() + "." + outputType);
		if (outputInfo.isDefault()) {
			candidates.add(factory.getId() + ".default." + scope.label);
			candidates.add(factory.getId() + ".default");
			candidates.add(factory.getId() + "." + scope.label);
			candidates.add(factory.getId());
		}
		String parent = factory.getParent();
		while (CommonUtils.isValid(parent)) {
			candidates.add(parent + "." + outputType + "." + scope.label);
			candidates.add(parent + "." + outputType);
			String newParent = factoryRegistry.getParentFactory(parent);
			
			// exclude cyclic references
			if (parent.equalsIgnoreCase(newParent))
				break;
			
			parent = newParent;
		}
		
		log.debug("GENERATING for factory [" + factory.getId() + "] :: CANDIDATES : ["
					+ candidates.stream().collect(Collectors.joining(";"))
					+ "]"
				);
		
		for (String candidate: candidates) {
			Template t = templateService.getTemplateById(candidate); 
			if (t != null) {
				log.debug("GENERATING for factory [" + factory.getId() + "] :: FOUND TEMPLATE : ["
						+ candidate	+ "]");
				Generated result = GeneratorFactory.createGenerator(t).generate(t, parameters);
				return result;
			} else {
				log.debug("GENERATING for factory [" + factory.getId() + "] :: NOT FOUND TEMPLATE : ["
						+ candidate	+ "]");
			}
		}
		return null;
	}

	public TemplateService getTemplateService() {
		return templateService;
	}

	public void setTemplateService(TemplateService templateService) {
		this.templateService = templateService;
	}
	
}

==> 1.11.0.zip: entaxy-public/platform/runtime/base/objects-base/object-factory/src/main/java/ru/entaxy/platform/base/objects/factory/impl/GenerationHelper.java <==
/*-
 * ~~~~~~licensing~~~~~~
 * test-producers
 * ==========
 * Copyright (C) 2020 - 2025 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.base.objects.factory.impl;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.osgi.service.component.annotations.CollectionType;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.osgi.service.component.annotations.ReferencePolicy;
import org.osgi.service.component.annotations.ReferencePolicyOption;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.base.generator.template.Template;
import ru.entaxy.base.generator.template.TemplateService;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generated;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generator;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.factory.GeneratorFactory;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory;
import ru.entaxy.platform.base.objects.factory.EntaxyFactoryRegistry;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.OutputInfo;
import ru.entaxy.platform.base.objects.factory.EntaxyFactory.SCOPE;
import ru.entaxy.platform.base.support.CommonUtils;

@Component(immediate = true, service = {GenerationHelper.class})
public class GenerationHelper {

	private static final Logger log = LoggerFactory.getLogger(GenerationHelper.class);

	@Reference(cardinality = ReferenceCardinality.MANDATORY, collectionType = CollectionType.SERVICE, policy = ReferencePolicy.DYNAMIC, policyOption = ReferencePolicyOption.GREEDY)
	protected volatile TemplateService templateService;
	
	@Reference(cardinality = ReferenceCardinality.MANDATORY)
	EntaxyFactoryRegistry factoryRegistry;
	
	public Generated generateForFactory(EntaxyFactory factory, String outputType, SCOPE scope, Map<String, Object> parameters) throws Exception {
		
		log.debug("GENERATING for factory [" + factory.getId() + "]");
		log.debug("GENERATING for factory [" + factory.getId() + "] :: PARAMETERS \n"
				 + parameters.entrySet().stream().map(e -> e.getKey() + ":" + (null==e.getValue()?"null":e.getValue()))
				 .collect(Collectors.toList())
				 .stream().collect(Collectors.joining(", "))
				);
		
		OutputInfo outputInfo = factory.getOutputByType(outputType);
		
		if (outputInfo == null) {
			// @TODO throw Exception
			log.debug("GENERATING for factory [" + factory.getId() + "] :: output [{}] not found", outputType);
			return null;
		}
			
		if (!outputInfo.getScopes().contains(scope.label)) {
			// @TODO throw Exception
			log.debug("GENERATING for factory [" + factory.getId() + "] :: scope [{}] not supported", scope.label);
			return null;
		}
			
		Map<String, Object> generationProperties = new HashMap<>();
		generationProperties.put(EntaxyFactory.GENERATION.ATTRIBUTES.OUTPUT, outputType);
		generationProperties.put(EntaxyFactory.GENERATION.ATTRIBUTES.SCOPE, scope.label);
		generationProperties.put(EntaxyFactory.GENERATION.ATTRIBUTES.FACTORY, factory.getId());
		
		// try to generate via GeneratorFactory
		log.debug("GENERATING for factory [" + factory.getId() + "] :: TRY via GeneratorFactory");
		
		Map<String, Object> generatorProperties = new HashMap<>(outputInfo.getConfig());
		generatorProperties.putAll(generationProperties);
		Generator g = GeneratorFactory.createGenerator(outputInfo.getGenerator(), generatorProperties);
		
		Map<String, Object> generationParameters = new HashMap<>(parameters);
		generationParameters.putAll(generationProperties);
		
		if (g.isGeneratable(generationParameters)) {
			log.debug("GENERATING for factory [" + factory.getId() + "] :: generating");
			return g.generate(generationParameters);
		} else {
			log.debug("GENERATING for factory [" + factory.getId() + "] :: NOT generatable");			
		}
		
		// try to generate via TemplateService
		log.debug("GENERATING for factory [" + factory.getId() + "] :: TRY via TemplateService");
		List<String> candidates = new LinkedList<>();
		candidates.add(factory.getId() + "." + outputType + "." + scope.label);
		candidates.add(factory.getId() + "." + outputType);
		if (outputInfo.isDefault()) {
			candidates.add(factory.getId() + ".default." + scope.label);
			candidates.add(factory.getId() + ".default");
			candidates.add(factory.getId() + "." + scope.label);
			candidates.add(factory.getId());
		}
		String parent = factory.getParent();
		while (CommonUtils.isValid(parent)) {
			candidates.add(parent + "." + outputType + "." + scope.label);
			candidates.add(parent + "." + outputType);
			String newParent = factoryRegistry.getParentFactory(parent);
			
			// exclude cyclic references
			if (parent.equalsIgnoreCase(newParent))
				break;
			
			parent = newParent;
		}
		
		log.debug("GENERATING for factory [" + factory.getId() + "] :: CANDIDATES : ["
					+ candidates.stream().collect(Collectors.joining(";"))
					+ "]"
				);
		
		for (String candidate: candidates) {
			Template t = templateService.getTemplateById(candidate); 
			if (t != null) {
				log.debug("GENERATING for factory [" + factory.getId() + "] :: FOUND TEMPLATE : ["
						+ candidate	+ "]");
				Generated result = GeneratorFactory.createGenerator(t).generate(t, parameters);
				return result;
			} else {
				log.debug("GENERATING for factory [" + factory.getId() + "] :: NOT FOUND TEMPLATE : ["
						+ candidate	+ "]");
			}
		}
		log.warn("Template not found for generation; factory={}, output={}, scope={}"
				, factory.getId()
				, outputType
				, scope.label);
		return null;
	}

	public TemplateService getTemplateService() {
		return templateService;
	}

	public void setTemplateService(TemplateService templateService) {
		this.templateService = templateService;
	}
	
}
